// Copyright Fuzamei Corp. 2020 All Rights Reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Package jsonclient 实现JSON rpc客户端请求功能
package client

import (
	"bytes"
	"crypto/tls"
	"encoding/json"
	"errors"
	"fmt"
	"github.com/golang/protobuf/proto"
	"io/ioutil"
	"net/http"
	"strconv"
	"strings"
	"time"

	sdk "gitlab.33.cn/link33/chain33-sdk-go"
	"gitlab.33.cn/link33/chain33-sdk-go/crypto"
	"gitlab.33.cn/link33/chain33-sdk-go/dapp/broker"
	"gitlab.33.cn/link33/chain33-sdk-go/dapp/storage"
	"gitlab.33.cn/link33/chain33-sdk-go/event"
	"gitlab.33.cn/link33/chain33-sdk-go/types"
)

// JSONClient a object of jsonclient
type JSONClient struct {
	url       string
	prefix    string
	tlsVerify bool
	client    *http.Client
}

func addPrefix(prefix, name string) string {
	if strings.Contains(name, ".") {
		return name
	}
	return prefix + "." + name
}

// NewJSONClient produce a json object
func NewJSONClient(prefix, url string) (*JSONClient, error) {
	return new(prefix, url, false)
}

// New produce a jsonclient by perfix and url
func new(prefix, url string, tlsVerify bool) (*JSONClient, error) {
	httpcli := http.DefaultClient
	if strings.Contains(url, "https") { // 暂不校验tls证书
		httpcli = &http.Client{Transport: &http.Transport{TLSClientConfig: &tls.Config{InsecureSkipVerify: !tlsVerify}}}
	}
	return &JSONClient{
		url:       url,
		prefix:    prefix,
		tlsVerify: tlsVerify,
		client:    httpcli,
	}, nil
}

type clientRequest struct {
	Method string         `json:"method"`
	Params [1]interface{} `json:"params"`
	ID     uint64         `json:"id"`
}

type clientResponse struct {
	ID     uint64           `json:"id"`
	Result *json.RawMessage `json:"result"`
	Error  interface{}      `json:"error"`
}

func (client *JSONClient) Call(method string, params, resp interface{}) error {
	method = addPrefix(client.prefix, method)
	req := &clientRequest{}
	req.Method = method
	req.Params[0] = params
	data, err := json.Marshal(req)
	if err != nil {
		return err
	}
	postresp, err := client.client.Post(client.url, "application/json", bytes.NewBuffer(data))
	if err != nil {
		return err
	}
	defer postresp.Body.Close()
	b, err := ioutil.ReadAll(postresp.Body)
	if err != nil {
		return err
	}
	cresp := &clientResponse{}
	err = json.Unmarshal(b, &cresp)
	if err != nil {
		return err
	}
	if cresp.Error != nil {
		x, ok := cresp.Error.(string)
		if !ok {
			return fmt.Errorf("invalid error %v", cresp.Error)
		}
		if x == "" {
			x = "unspecified error"
		}
		return fmt.Errorf(x)
	}
	if cresp.Result == nil {
		return errors.New("Empty result")
	}
	if msg, ok := resp.(proto.Message); ok {
		var str json.RawMessage
		err = json.Unmarshal(*cresp.Result, &str)
		if err != nil {
			return err
		}
		b, err := str.MarshalJSON()
		if err != nil {
			return err
		}
		err = json.Unmarshal(b, msg)
		if err != nil {
			fmt.Println("err", err)
			return err
		}
		return nil
	}
	return json.Unmarshal(*cresp.Result, resp)
}

type ParseFunc func(result json.RawMessage) (interface{}, error)

// 回调函数，用于自定义解析返回得result数据
func (client *JSONClient) CallBack(method string, params interface{}, parseFunc ParseFunc) (interface{}, error) {
	method = addPrefix(client.prefix, method)
	req := &clientRequest{}
	req.Method = method
	req.Params[0] = params
	data, err := json.Marshal(req)
	if err != nil {
		return nil, err
	}
	postresp, err := client.client.Post(client.url, "application/json", bytes.NewBuffer(data))
	if err != nil {
		return nil, err
	}
	defer postresp.Body.Close()
	b, err := ioutil.ReadAll(postresp.Body)
	if err != nil {
		return nil, err
	}
	cresp := &clientResponse{}
	err = json.Unmarshal(b, &cresp)
	if err != nil {
		return nil, err
	}
	if cresp.Error != nil {
		x, ok := cresp.Error.(string)
		if !ok {
			return nil, fmt.Errorf("invalid error %v", cresp.Error)
		}
		if x == "" {
			x = "unspecified error"
		}
		return nil, fmt.Errorf(x)
	}
	if cresp.Result == nil {
		return nil, errors.New("Empty result")
	}
	return parseFunc(*cresp.Result)
}

// 发送交易
func (client *JSONClient) SendTransaction(signedTx string) (string, error) {
	var res string
	send := &RawParm{
		Token: "BTY",
		Data:  signedTx,
	}
	err := client.Call("Chain33.SendTransaction", send, &res)
	if err != nil {
		return "", err
	}

	return res, nil
}

// SendTransactionSync
func (client *JSONClient) SendTransactionSync(privkey string, tx *types.Transaction) (*Response, error) {
	// 签名
	hexbytes, _ := types.FromHex(privkey)
	sdk.Sign(tx, hexbytes, crypto.SECP256K1, nil)
	signedTx := types.ToHexPrefix(types.Encode(tx))
	var res string
	send := &RawParm{
		Token: "BTY",
		Data:  signedTx,
	}
	err := client.Call("Chain33.SendTransactionSync", send, &res)
	if err != nil {
		return nil, err
	}
	return &Response{OK: true, Message: res, Data: types.Encode(tx)}, nil
}

// 查询交易
func (client *JSONClient) QueryTransaction(hash string) (*TransactionDetail, error) {
	query := QueryParm{
		Hash: hash,
	}
	var detail TransactionDetail
	err := client.Call(Chain33_QueryTransaction, query, &detail)
	if err != nil {
		return nil, err
	}

	return &detail, nil
}

func (client *JSONClient) QueryLastHeader() (*Header, error) {
	var res Header
	err := client.Call(Chain33_GetLastHeader, nil, &res)
	if err != nil {
		return nil, err
	}
	return &res, nil
}

func (client *JSONClient) QueryBlockHash(height int64) (*ReplyHash, error) {
	var res ReplyHash
	params := types.ReqInt{
		Height: height,
	}
	err := client.Call(Chain33_GetBlockHash, params, &res)
	if err != nil {
		return nil, err
	}
	return &res, nil
}

func (client *JSONClient) QueryHeadersByHeight(height int64) (*Headers, error) {
	var res Headers
	params := BlockParam{
		Start:    height - 1,
		End:      height - 1,
		Isdetail: false,
	}
	err := client.Call(Chain33_GetHeaders, params, &res)
	if err != nil {
		return nil, err
	}
	return &res, nil
}

func (client *JSONClient) QueryIsSync() (bool, error) {
	var res bool
	err := client.Call(Para_IsSync, nil, &res)
	if err != nil {
		return false, err
	}
	return res, nil
}

func (client *JSONClient) QueryMainNetIsSync() (bool, error) {
	var res bool
	err := client.Call(Chain33_IsSync, nil, &res)
	if err != nil {
		return false, err
	}
	return res, nil
}

func (client *JSONClient) QueryNetInfo() (*NodeNetinfo, error) {
	var res NodeNetinfo
	params := types.ReqNil{}
	err := client.Call(Chain33_GetNetInfo, params, &res)
	if err != nil {
		return nil, err
	}
	return &res, nil
}

func (client *JSONClient) QueryMempool(jrpcURl string) (*ReplyTxList, error) {
	var res ReplyTxList
	params := types.ReqNil{}
	err := client.Call(Chain33_GetMempool, params, &res)
	if err != nil {
		return nil, err
	}
	return &res, nil
}

// 获取区块信息
func (client *JSONClient) QueryBlockInfo(start, end int64, isDetail bool) (*BlockDetails, error) {
	var res BlockDetails
	params := BlockParam{Start: start, End: end, Isdetail: isDetail}
	err := client.Call(Chain33_GetBlocks, params, &res)
	if err != nil {
		return nil, err
	}
	return &res, nil
}

// k V明文存证
func (client *JSONClient) KVStore(key, value, privateKey string) (*Response, error) {
	tx, err := storage.CreateContentStorageTx("", storage.OpCreate, key, nil, value)
	if err != nil {
		return nil, err
	}
	return client.SendTransactionSync(privateKey, tx)
}

// 初始化broker
func (client *JSONClient) InitBroker(bxhId, appChainId, privateKey string) (*Response, error) {
	tx := broker.InitBroker(bxhId, appChainId)
	return client.SendTransactionSync(privateKey, tx)
}

// 发布跨链事件
func (client *JSONClient) EmitDataSwapInterchainEvent(dstServiceID, args, argscb, argsrb string, privateKey string) (*Response, error) {
	// FIXME 这里需完善     funcName="interchainGet,interchainSet," txType=0 表示是存在类数据交换
	tx := broker.EmitInterchainEvent("", dstServiceID, "interchainGet,interchainSet,", args, argsrb, argsrb, 0)
	return client.SendTransactionSync(privateKey, tx)
}

// TODO 执行请求
func (client *JSONClient) Execute(requst *Request, privateKey string) (*Response, error) {
	if requst.Exec == broker.BrokerX {
		switch requst.Fcn {
		case "invokeInterchain":
			srcChainServiceID := string(requst.Args[0])
			sequenceNum, err := strconv.ParseUint(string(requst.Args[1]), 10, 64)
			if err != nil {
				return nil, fmt.Errorf("cannot parse %s to uint64", string(requst.Args[1]))
			}
			targetCID := string(requst.Args[2])
			reqType, err := strconv.ParseUint(string(requst.Args[3]), 10, 64)
			if err != nil {
				return nil, fmt.Errorf("cannot parse %s to uint64", string(requst.Args[3]))
			}

			// reqType==0 表示跨入交易， reqType==1 表示跨出交易，响应更新
			if reqType != 0 {
				tx := broker.UpdateIndex("", srcChainServiceID, targetCID, sequenceNum, reqType, 0)
				resp, err := client.SendTransactionSync(privateKey, tx)
				if err != nil {
					return resp, err
				}
			}

			//splitedCID[1]就是合约名称
			splitedCID := strings.Split(targetCID, "&")
			if len(splitedCID) != 2 {
				return nil, fmt.Errorf("Target chaincode id %s is not valid", targetCID)
			}

			callFunc := &CallFunc{}
			if err := json.Unmarshal(requst.Args[4], callFunc); err != nil {
				return nil, fmt.Errorf("unmarshal call func failed for %s", requst.Args[4])
			}
			fmt.Println("==================callfunc=======",callFunc.Func)
			// 调用相应链码， （链码名称，参数，通道名称）
			if splitedCID[1] == storage.StorageX || splitedCID[1] == broker.BrokerX {
				//interchainGet方法是在sdk封装的，用于支持跨链之前数据的查询
				switch callFunc.Func {
				case "interchainGet":
					response, err := client.InterChainGet(string(callFunc.Args[0]))
					if err != nil {
						return response, err
					}
					//携带TXhash
					//response.Message = resp.Message
					return response, err
				case "interchainSet":
					//跨链set kv
					status := 1
					response, err := client.InterChainSet(string(callFunc.Args[0]), string(callFunc.Args[1]), privateKey)
					if err != nil {
						status = 0
					}
					//TODO 这里的处理不是事务机制
					if reqType != 0 {
						tx := broker.UpdateIndex("", srcChainServiceID, targetCID, sequenceNum, reqType, int32(status))
						resp, err := client.SendTransactionSync(privateKey, tx)
						if err != nil {
							return resp, err
						}
					}
					return response, nil
				}
			} else {
				return nil, fmt.Errorf("can't support exec %s", splitedCID[1])
			}

		}

	}

	return nil, fmt.Errorf("not matching execution or method!")
}

//包装适配跨链插件调用
func (client *JSONClient) InterChainSet(key, value, privateKey string) (*Response, error) {
	tx, err := storage.CreateContentStorageTx("", 0, key, nil, value)
	if err != nil {
		return nil, err
	}
	resp, err := client.SendTransactionSync(privateKey, tx)
	return resp, err
}

//包装适配跨链插件调用
func (client *JSONClient) InterChainGet(key string) (*Response, error) {
	jsonraw, err := json.Marshal(&types.QueryStorage{TxHash: key})
	if err != nil {
		return nil, err
	}
	query := Query4Jrpc{
		Execer:   client.prefix + storage.StorageX,
		FuncName: storage.FuncNameQueryStorage,
		Payload:  jsonraw,
	}
	// var storage types.Storage
	data, err := client.CallBack("Chain33.Query", query, ParseStorage)
	if err != nil {
		return nil, err
	}
	storage := data.(*types.Storage)
	//这里暂时就去明文存证内容
	resp := &Response{OK: true, Message: key, Data: []byte(storage.GetContentStorage().GetValue())}
	return resp, nil
}

// 简单实现交易事件的监听功能,后面换成grpc长链接方式监听
func (client *JSONClient) RegisterTxEvent(start int64, ccID, eventFilter string) (chan<- event.Registration, <-chan *event.CCEvent, <-chan error) {
	ch := make(chan *event.CCEvent, 1)
	resCh := make(chan event.Registration, 1)
	errCh := make(chan error, 1)
	// FIXME 这里实现是以联盟链方式，联盟链没有回滚，所以这里暂时按高度拉去处理
	go func(ch chan<- *event.CCEvent, closeCh <-chan event.Registration, errCh chan<- error) {
		height := start
	HERE:
		header, err := client.QueryLastHeader()
		if err != nil {
			errCh <- err
			return
		}
		lastHeight := header.Height
		if lastHeight <= height {
			time.Sleep(time.Second)
			goto HERE
		}
		ticket := time.After(100 * time.Millisecond)

		select {
		case <-ticket:
			blockInfo, err := client.QueryBlockInfo(height, height, true)
			if err != nil {
				errCh <- err
				return
			}

			for _, item := range blockInfo.Items {
				for i, tx := range item.Block.Txs {
					if tx.Execer == ccID && item.Receipts[i].Ty == types.ExecOk {
						// todo 这里需要重新弄成通用处理
						if tx.Execer == broker.BrokerX {
							var brokerAction types.BrokerAction
							err := types.Decode(tx.Payload, &brokerAction)
							if err != nil {
								errCh <- err
								return
							}
							if brokerAction.Ty == broker.ActionMap[eventFilter] {
								ccEvent := &event.CCEvent{Payload: types.Encode(brokerAction.GetEmitInterchainEvent()), CCID: tx.Execer, TxID: tx.Hash, EventName: eventFilter, BlockNumber: uint64(item.Block.Height)}
								ch <- ccEvent
							}
						}
					}
				}
			}
			height++
			goto HERE

		case <-closeCh:
			break
		}

	}(ch, resCh, errCh)
	return resCh, ch, nil
}

func (client *JSONClient) PollingEvent(meta *types.Meta) (*types.InterChainEventList, error) {
	jsonraw, err := json.Marshal(&types.PollingEvent{Meta: meta})
	if err != nil {
		return nil, err
	}
	query := Query4Jrpc{
		Execer:   client.prefix + broker.BrokerX,
		FuncName: broker.FuncNamePollingEvent,
		Payload:  jsonraw,
	}
	var eventList types.InterChainEventList
	err = client.Call("Chain33.Query", query, &eventList)
	if err != nil {
		return nil, err
	}
	return &eventList, nil
}

func (client *JSONClient) QueryInnerMeta() (*types.Meta, error) {
	jsonraw, err := json.Marshal(&types.QueryInnerMeta{})
	if err != nil {
		return nil, err
	}
	query := Query4Jrpc{
		Execer:   client.prefix + broker.BrokerX,
		FuncName: broker.FuncNameQueryInnerMeta,
		Payload:  jsonraw,
	}
	//var meta types.Meta
	//err = client.Call("Chain33.Query", query, &meta)
	//if err != nil {
	//	return nil, err
	//}
	//return &meta, nil
	data, err := client.CallBack("Chain33.Query", query, ParseMeta)
	if err != nil {
		return nil, err
	}
	return data.(*types.Meta), nil
}

func (client *JSONClient) QueryOutterMeta() (*types.Meta, error) {
	jsonraw, err := json.Marshal(&types.QueryOutterMeta{})
	if err != nil {
		return nil, err
	}
	query := Query4Jrpc{
		Execer:   client.prefix + broker.BrokerX,
		FuncName: broker.FuncNameQueryOutterMeta,
		Payload:  jsonraw,
	}
	//var meta types.Meta
	//err = client.Call("Chain33.Query", query, &meta)
	//if err != nil {
	//	return nil, err
	//}
	//return &meta, nil
	data, err := client.CallBack("Chain33.Query", query, ParseMeta)
	if err != nil {
		return nil, err
	}
	return data.(*types.Meta), nil
}

func (client *JSONClient) QueryCallBackMeta() (*types.Meta, error) {
	jsonraw, err := json.Marshal(&types.QueryNilParam{})
	if err != nil {
		return nil, err
	}
	query := Query4Jrpc{
		Execer:   client.prefix + broker.BrokerX,
		FuncName: broker.FuncNameQueryCallBackMeta,
		Payload:  jsonraw,
	}
	var meta types.Meta
	err = client.Call("Chain33.Query", query, &meta)
	if err != nil {
		return nil, err
	}
	return &meta, nil
}

func (client *JSONClient) QueryOutMessage(inServicePair string, index uint64) (*types.InterchainEvent, error) {
	jsonraw, err := json.Marshal(&types.QueryOutMessage{InServicePair: inServicePair, SequenceNum: index})
	if err != nil {
		return nil, err
	}
	query := Query4Jrpc{
		Execer:   client.prefix + broker.BrokerX,
		FuncName: broker.FuncNameQueryOutMessage,
		Payload:  jsonraw,
	}
	//var event types.InterchainEvent
	//err = client.Call("Chain33.Query", query, &event)
	//if err != nil {
	//	return nil, err
	//}
	//return &event, nil
	data, err := client.CallBack("Chain33.Query", query, ParseInterChainEvent)
	if err != nil {
		return nil, err
	}
	return data.(*types.InterchainEvent), nil
}

func (client *JSONClient) QueryInMessage(inServicePair string, index uint64) (*types.Response, error) {
	jsonraw, err := json.Marshal(&types.QueryInMessage{InServicePair: inServicePair, SequenceNum: index})
	if err != nil {
		return nil, err
	}
	query := Query4Jrpc{
		Execer:   client.prefix + broker.BrokerX,
		FuncName: broker.FuncNameQueryInMessage,
		Payload:  jsonraw,
	}
	var resp types.Response
	err = client.Call("Chain33.Query", query, &resp)
	if err != nil {
		return nil, err
	}
	return &resp, nil
}

func (client *JSONClient) QueryBrokerInfo() (*types.BrokerInfo, error) {
	jsonraw, err := json.Marshal(&types.QueryNilParam{})
	if err != nil {
		return nil, err
	}
	query := Query4Jrpc{
		Execer:   client.prefix + broker.BrokerX,
		FuncName: broker.FuncNameQueryBrokerInfo,
		Payload:  jsonraw,
	}
	var resp types.BrokerInfo
	err = client.Call("Chain33.Query", query, &resp)
	if err != nil {
		return nil, err
	}
	return &resp, nil
}
